import asyncio

from brigid.application.application import with_app
from brigid.cli.application import app
from brigid.core import logging
from brigid.library.storage import storage
from brigid.renderer.markdown_render import render_page

logger = logging.get_module_logger()


def is_list_line(line: str) -> bool:
    if line.startswith("**"):
        return False

    line = line.lstrip()

    return line.startswith("-") or line.startswith("*") or line.startswith("+")


def is_line_empty(line: str) -> bool:
    return not line.strip()


# TODO: what with nested lists?
# TODO: what with numbered lists?
# TODO: make difference for different list types (-, *, +, digit)
def parser(text: str) -> bool:  # noqa: CCR001

    is_in_list = False

    can_has_problem = False

    for line in text.split("\n"):
        if is_list_line(line):
            if can_has_problem:
                return True

            is_in_list = True
            continue

        if is_line_empty(line) and is_in_list:
            can_has_problem = True
            continue

        is_in_list = False
        can_has_problem = False

    return False


async def run() -> None:
    errors = []

    async with with_app():
        # pages are loaded on the application initialization

        site = storage.get_site()

        for page in storage.all_pages():
            context = render_page(page=page)

            if parser(page.body):
                context.add_error("test is unknown", "the page has separated list items")

            errors.extend(context.errors)

        article_404 = storage.get_article(slug="404")

        if set(article_404.pages) != set(site.allowed_languages):
            raise ValueError("404 article pages are not consistent with site allowed languages")

        article_500 = storage.get_article(slug="500")

        if set(article_500.pages) != set(site.allowed_languages):
            raise ValueError("500 article pages are not consistent with site allowed languages")

    if not errors:
        logger.info("everything_is_ok")
        return

    for error in errors:
        logger.error(
            "markdown_render_error",
            filepath=str(error.filepath),
            failed_text=error.failed_text,
            message=error.message,
        )

    logger.error("markdown_render_errors", errors_number=len(errors))


@app.command()
def validate() -> None:
    asyncio.run(run())



from typing import Any, TypeVar

import pydantic

BASE_ENTITY = TypeVar("BASE_ENTITY", bound="BaseEntity")


class BaseEntity(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        str_strip_whitespace=True,
        validate_default=True,
        extra="forbid",
        frozen=True,
        validate_assignment=True,
        from_attributes=False,
    )

    def replace(self: BASE_ENTITY, **kwargs: Any) -> BASE_ENTITY:
        data = self.model_dump()
        data |= kwargs
        return self.__class__(**data)



import re

from markdown.inlinepatterns import LINK_RE as INTERNAL_LINK_RE
from markdown.inlinepatterns import LinkInlineProcessor

from brigid.domain.urls import UrlsFeedsAtom, UrlsPost, UrlsTags, normalize_url
from brigid.library.storage import storage
from brigid.renderer.context import render_context


class InternalLinkInlineProcessor(LinkInlineProcessor):
    RE_LINK = re.compile(r"\{\s*(.*?)\s*\}", re.DOTALL | re.UNICODE)

    # TODO: return errors instead of result
    def handleMatch(self, m, data):  # noqa # pylint: disable=all
        from brigid.library.connectivity import connectivity

        result = super().handleMatch(m, data)

        if result[0] is None:
            return result

        href = result[0].get("href")

        if href is None:
            return result

        site = storage.get_site()

        context = render_context.get()

        parts = href.split(":")

        new_href = None

        # TODO: support specifying language to easier give cross-language links

        if parts[0] == "post":
            slug = parts[1]

            if not storage.has_article(slug=slug):
                context = render_context.get()

                context.add_error(
                    failed_text=data,
                    message="Non-existing page slug specified for local link",
                )
                return result

            article = storage.get_article(slug=slug)

            if context.page.language not in article.pages:
                context.add_error(
                    failed_text=data,
                    message="Article does not have page in the current language",
                )
                return result

            connectivity.add_connection(
                target_page_id=article.pages[context.page.language],
                reference_page_id=context.page.id,
            )

            new_href = UrlsPost(language=context.page.language, slug=slug).url()

        elif parts[0] == "tags":

            required = set()
            excluded = set()

            for tag in parts[1:]:
                normalized_tag = tag[1:] if tag[0] == "-" else tag

                if normalized_tag not in site.languages[context.page.language].tags_translations:
                    context.add_error(failed_text=data, message="One of tags does not exist")
                    return result

                if tag[0] != "-":
                    required.add(normalized_tag)
                else:
                    excluded.add(normalized_tag)

            new_href = UrlsTags(
                page=1,
                required_tags=required,
                excluded_tags=excluded,
                language=context.page.language,
            ).url()

        elif parts[0] == "absolute":

            if parts[1][0] != "/":
                context.add_error(failed_text=data, message="Absolute link should start with /")
                return result

            new_href = normalize_url(f"{site.url}/{parts[1]}")

        elif parts[0] == "feed":
            new_href = UrlsFeedsAtom(language=context.page.language).url()

        elif parts[0] == "static":
            post_url = UrlsPost(language=context.page.language, slug=context.article.slug)

            new_href = post_url.file_url(parts[1])

        else:
            context.add_error(failed_text=data, message="Unknown local link format")
            return result

        # set full url for rendering in feeds and other places
        result[0].set("href", new_href)

        return result

    def getLink(self, data: str, index: int) -> tuple[str, str | None, int, bool]:
        match = self.RE_LINK.match(data, index)

        if match is None:
            return "", None, -1, False

        href = match.group(1)

        return href, None, match.end(), True


__all__ = ["InternalLinkInlineProcessor", "INTERNAL_LINK_RE"]



from html.parser import HTMLParser


class MLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.fed = []

    def handle_data(self, data):
        self.fed.append(data)

    def get_data(self):
        return "".join(self.fed)


def strip_html(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()



from typing import Any

import pydantic
from pymdownx.blocks import BlocksExtension

from brigid.core.entities import BaseEntity
from brigid.renderer.context import render_context
from brigid.renderer.processors.toml_block import TomlBlock


class CollectionModel(BaseEntity):
    id: str
    css_class: str | None = None

    model_config = pydantic.ConfigDict(frozen=False)


class CollectionBlock(TomlBlock):
    NAME = "brigid-collection"
    models = CollectionModel
    root_tag = "div"

    def root_css_classes(self, data: Any) -> list[str]:
        classes = ["brigid-collection", data.css_class]
        return [x for x in classes if x]

    def process_data(self, data: Any) -> str:
        from brigid.theme.templates import render

        context = render_context.get()

        return render(
            "./blocks/collection.html.j2",
            {
                "collection_data": data,
                # TODO: rename all arguments in all blogs to "current_"
                # TODO: add current_page everywhere
                "current_article": context.article,
                "current_page": context.page,
            },
        )


class CollectionBlockExtension(BlocksExtension):

    def extendMarkdownBlocks(self, md, block_mgr):
        block_mgr.register(CollectionBlock, self.getConfigs())



import contextlib
from typing import AsyncGenerator

import fastapi
from fastapi.middleware.cors import CORSMiddleware

from brigid.api import http_handlers as api_http_handlers
from brigid.api import middlewares as api_middlewares
from brigid.api import static_cache as api_static_cache
from brigid.api.settings import settings as api_settings
from brigid.application.settings import settings
from brigid.core import logging, sentry
from brigid.library import discovering
from brigid.library.settings import settings as library_settings

logger = logging.get_module_logger()


def initialize_api(app: fastapi.FastAPI) -> None:
    logger.info("initialize_api")

    app.include_router(api_http_handlers.router)

    app.middleware("http")(api_middlewares.remove_double_slashes)
    app.middleware("http")(api_middlewares.remove_trailing_slash)
    app.middleware("http")(api_middlewares.set_content_language)

    app.exception_handler(404)(api_middlewares.process_404)
    app.exception_handler(Exception)(api_middlewares.process_expected_error)

    if api_settings.cache_directory:
        cache = api_static_cache.FileCache(directory=api_settings.cache_directory)
        api_static_cache.set_cache(cache)

    logger.info("api_initialized")


@contextlib.asynccontextmanager
async def use_sentry() -> AsyncGenerator[None, None]:
    logger.info("sentry_enabled")

    sentry.initialize(
        dsn=settings.sentry.dsn,
        sample_rate=settings.sentry.sample_rate,
        traces_sample_rate=settings.sentry.traces_sample_rate,
        environment=settings.environment,
    )

    logger.info("sentry_initialized")

    yield

    logger.info("sentry_disabled")


def smart_url(domain: str, port: int) -> str:
    if port == 80:
        return f"http://{domain}"

    if port == 443:
        return f"https://{domain}"

    return f"http://{domain}:{port}"


def create_app() -> fastapi.FastAPI:  # noqa: CCR001
    logging.initialize(use_sentry=settings.sentry.enabled)

    logger.info("create_app")

    @contextlib.asynccontextmanager
    async def lifespan(app: fastapi.FastAPI) -> AsyncGenerator[None, None]:
        async with contextlib.AsyncExitStack() as stack:
            if settings.sentry.enabled:
                await stack.enter_async_context(use_sentry())

            await app.router.startup()

            discovering.load(directory=library_settings.directory)

            yield

            await app.router.shutdown()

    app = fastapi.FastAPI(
        lifespan=lifespan,
        docs_url=None,
        redoc_url=None,
        openapi_url=None,
    )

    initialize_api(app)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=settings.origins,
        allow_credentials=False,
        allow_methods=[],
        allow_headers=[],
    )

    logger.info("app_created")

    return app


@contextlib.asynccontextmanager
async def with_app() -> AsyncGenerator[fastapi.FastAPI, None]:
    async with app.router.lifespan_context(app):
        yield app


app = create_app()



import pathlib
from typing import Any

import pydantic
from pymdownx.blocks import BlocksExtension

from brigid.core.entities import BaseEntity
from brigid.domain.html import strip_html
from brigid.library.entities import Article
from brigid.renderer.context import render_context
from brigid.renderer.processors.toml_block import TomlBlock

# TODO: remove duplicated code


class ImageMixing:

    @pydantic.validator("alt")
    def escape_alt(cls, v):
        if v is None:
            return v

        return strip_html(v)


class Image(ImageMixing, BaseEntity):
    src: str
    alt: str | None = None

    # TODO: add caption to the image to show in the lightbox

    model_config = pydantic.ConfigDict(frozen=False)

    def url_file(self, article: Article) -> pathlib.Path:
        return article.path.parent / self.src


class ImagesModel(BaseEntity):
    caption: str | None = None
    images: list[Image]

    # TODO: rename to css_class
    galery_class: str | None = None

    model_config = pydantic.ConfigDict(frozen=False)

    @pydantic.validator("images")
    def minimum_1_image(cls, v):
        if len(v) < 1:
            raise ValueError("at least one image is required")
        return v

    @pydantic.model_validator(mode="after")
    def first_image_alt_from_caption(cls, self):

        if self.caption is not None and self.images[0].alt is None:
            self.images[0].alt = self.caption

        for image in self.images:
            if image.alt is None:
                raise ValueError("alt must be present")

        return self

    @pydantic.model_validator(mode="after")
    def default_galery_class(cls, self):
        if self.galery_class is None:
            self.galery_class = f"brigid-images-{len(self.images)}"

        return self


class ImageModel(ImageMixing, BaseEntity):
    src: str
    alt: str | None = None
    caption: str | None = None
    galery_class: str = "brigid-images-1"

    model_config = pydantic.ConfigDict(frozen=False)

    @pydantic.model_validator(mode="after")
    def alt_or_caption(cls, self):
        if self.alt is None and self.caption is None:
            raise ValueError("alt or caption must be present")

        if self.alt is None and self.caption is not None:
            self.alt = self.caption

        return self


class ImagesBlock(TomlBlock):
    NAME = "brigid-images"
    models = ImageModel | ImagesModel
    root_tag = "figure"

    def root_css_classes(self, data: Any) -> list[str]:
        return ["brigid-images", data.galery_class]

    def process_data(self, data: Any) -> str:
        from brigid.theme.templates import render

        context = render_context.get()

        if isinstance(data, ImageModel):
            images = ImagesModel(
                caption=data.caption,
                images=[Image(src=data.src, alt=data.alt)],
                galery_class=data.galery_class,
            )
        else:
            images = data

        return render(
            "./blocks/images.html.j2",
            {
                "images": images,
                "article": context.article,
                "page": context.page,
            },
        )


class ImagesBlockExtension(BlocksExtension):

    def extendMarkdownBlocks(self, md, block_mgr):
        block_mgr.register(ImagesBlock, self.getConfigs())



import pathlib

from PIL import Image

from brigid.core.entities import BaseEntity


class ImageInfo(BaseEntity):
    path: pathlib.Path
    width: int
    height: int


class FilesInfo:
    __slots__ = ("_images",)

    def __init__(self) -> None:
        self._images: dict[pathlib.Path, ImageInfo] = {}

    def image_info(self, path: pathlib.Path) -> ImageInfo:
        if path in self._images:
            return self._images[path]

        with Image.open(path) as img:
            width, height = img.size

        self._images[path] = ImageInfo(path=path, width=width, height=height)

        return self._images[path]


files = FilesInfo()



import pathlib
from typing import Iterable

from brigid.library.entities import Article, Collection, Page, Redirects, Site


class Storage:
    __slots__ = (
        "_pages",
        "_articles",
        "_articles_by_path",
        "_articles_by_slug",
        "_site",
        "_redirects",
        "_collections",
    )

    def __init__(self) -> None:
        self._pages: dict[str, Page] = {}
        self._articles: dict[str, Article] = {}
        self._collections: dict[str, Collection] = {}

        self._articles_by_path: dict[pathlib.Path, str] = {}
        self._articles_by_slug: dict[str, str] = {}

        self._site: Site | None = None
        self._redirects: Redirects | None = None

    def get_article(
        self,
        id: str | None = None,
        path: pathlib.Path | None = None,
        slug: str | None = None,
    ) -> Article:

        if sum(1 for x in (id, path, slug) if x is not None) != 1:
            raise NotImplementedError("id, path and slug are mutually exclusive")

        if id is not None:
            return self._articles[id]

        if path is not None:
            return self._articles[self._articles_by_path[path]]

        if slug is not None:
            return self._articles[self._articles_by_slug[slug]]

        raise NotImplementedError("One of id, path or slug must be provided")

    def has_article(
        self,
        id: str | None = None,
        path: pathlib.Path | None = None,
        slug: str | None = None,
    ) -> bool:

        if sum(1 for x in (id, path, slug) if x is not None) != 1:
            raise NotImplementedError("id, path and slug are mutually exclusive")

        if id is not None:
            return id in self._articles

        if path is not None:
            return path in self._articles_by_path

        if slug is not None:
            return slug in self._articles_by_slug

        raise NotImplementedError("One of id, path or slug must be provided")

    def add_article(self, article: Article, replace: bool = False) -> None:
        if not replace and article.id in self._articles:
            raise NotImplementedError("article with id already exists")

        self._articles[article.id] = article
        self._articles_by_path[article.path] = article.id
        self._articles_by_slug[article.slug] = article.id

    def add_page(self, page: Page, replace: bool = False) -> None:
        if page.article_id not in self._articles:
            raise NotImplementedError("article must be added first")

        article = self.get_article(id=page.article_id)

        if not replace and page.language in article.pages:
            raise NotImplementedError("page with language already exists")

        if not replace and page.id in article.pages.values():
            raise NotImplementedError("page with id already exists")

        article.pages[page.language] = page.id

        self._pages[page.id] = page

    def get_page(self, id: str) -> Page:
        return self._pages[id]

    def set_site(self, site: Site) -> None:
        self._site = site

    def get_site(self) -> Site:
        assert self._site
        return self._site

    def set_redirects(self, redirects: Redirects) -> None:
        self._redirects = redirects

    def get_redirects(self) -> Redirects:
        assert self._redirects
        return self._redirects

    def add_collection(self, collection: Collection) -> None:
        if collection.id in self._collections:
            raise NotImplementedError("collection already added")

        self._collections[collection.id] = collection

    def get_collection(self, id: str) -> Collection:
        return self._collections[id]

    def all_pages(self) -> list[Page]:
        return list(self._pages.values())

    def pages_number(self, language: str) -> int:
        return sum(1 for page in self._pages.values() if page.language == language)

    # TODO: cache
    def last_pages(
        self,
        language: str,
        number: int | None = None,
        only_posts: bool = True,
        require_tags: Iterable[str] = (),
        exclude_tags: Iterable[str] = (),
    ) -> list[Page]:
        pages = [
            page for page in self._pages.values() if page.language == language and (not only_posts or page.is_post)
        ]

        if require_tags:
            require_tags = set(require_tags)
            pages = [page for page in pages if page.tags >= require_tags]

        if exclude_tags:
            exclude_tags = set(exclude_tags)
            pages = [page for page in pages if not page.tags & exclude_tags]

        pages.sort(key=lambda x: x.published_at, reverse=True)

        if number is None:
            return pages

        return pages[:number]


storage = Storage()



import pathlib
import shutil


class BaseCache:
    __slots__ = ()

    def initialize(self) -> None:
        raise NotImplementedError("initialize")

    def clear(self) -> None:
        raise NotImplementedError("clear")

    def set(self, cache_path: str, original_path) -> None:
        raise NotImplementedError("set")


class DummyCache(BaseCache):
    __slots__ = ()

    def initialize(self) -> None:
        pass

    def clear(self) -> None:
        pass

    def set(self, cache_path: str, original_path) -> None:
        pass


class FileCache(BaseCache):
    __slots__ = ("directory",)

    def __init__(self, directory: pathlib.Path) -> None:
        super().__init__()
        self.directory = directory

    def initialize(self) -> None:
        if not self.directory.exists():
            self.directory.mkdir(parents=True)

        self.clear()

    def clear(self) -> None:
        for entry in self.directory.iterdir():
            if entry.is_dir():
                shutil.rmtree(entry)
            else:
                entry.unlink()

    def set(self, raw_cache_path: str, original_path) -> None:

        if raw_cache_path.startswith("/"):
            raw_cache_path = raw_cache_path[1:]

        cache_path = self.directory / raw_cache_path

        if not cache_path.parent.exists():
            cache_path.parent.mkdir(parents=True)

        shutil.copy(original_path, cache_path)


_cache: BaseCache = DummyCache()


def set_cache(cache: FileCache) -> None:
    cache.initialize()

    global _cache

    _cache = cache


def cache() -> BaseCache:
    assert _cache is not None
    return _cache



translations = {
    "ru": {
        "page": "страница",
        "with_tags": "с тегами",
        "without_tags": "без тегов",
    },
    "en": {
        "page": "page",
        "with_tags": "with tags",
        "without_tags": "without tags",
    },
}
